@@ -148,6 +148,7 @@ _ffi.cdef('''
     #define MDB_READERS_FULL ...
     #define MDB_REVERSEKEY ...
     #define MDB_TXN_FULL ...
+    #define MDB_WRITEMAP ...
 
     // Helpers below inline MDB_vals. Avoids key alloc/dup on CPython, where
     // cffi will use PyString_AS_STRING when passed as an argument.
@@ -336,7 +337,8 @@ class Environment(object):
 
         `readonly`:
             If ``True``, disallow any write operations. Note the lock file is
-            still modified.
+            still modified. If specified, the ``write`` flag to
+            :py:meth:`begin` or :py:class:`Transaction` is ignored.
 
         `metasync`:
             If ``False``, never explicitly flush metadata pages to disk. OS
@@ -355,6 +357,11 @@ class Environment(object):
         `create`:
             If ``False``, do not create the directory `path` if it is missing.
 
+        `writemap`:
+            If ``True`` LMDB will use a writeable memory map to update the
+            database. Requires operating system support. Please refer to
+            documentation.
+
         `max_readers`:
             Slots to allocate in lock file for read threads; attempts to open
             the environment by more than this many clients simultaneously will
@@ -366,7 +373,8 @@ class Environment(object):
     """
     def __init__(self, path, map_size=10485760, subdir=True,
             readonly=False, metasync=True, sync=True, map_async=False,
-            mode=0644, create=True, max_readers=126, max_dbs=0):
+            mode=0644, create=True, writemap=False, max_readers=126,
+            max_dbs=0):
         envpp = _ffi.new('MDB_env **')
 
         rc = mdb_env_create(envpp)
@@ -395,17 +403,20 @@ class Environment(object):
             flags |= MDB_NOSUBDIR
         if readonly:
             flags |= MDB_RDONLY
+        self.readonly = readonly
         if not metasync:
             flags |= MDB_NOMETASYNC
         if not sync:
             flags |= MDB_NOSYNC
         if map_async:
             flags |= MDB_MAPASYNC
+        if writemap:
+            flags |= MDB_WRITEMAP
 
         rc = mdb_env_open(self._env, path, flags, mode)
         if rc:
             raise Error(path, rc)
-        with self.begin(db=object(), write=True) as txn:
+        with self.begin(db=object()) as txn:
             self._db = _Database(self, txn, None, False, False, True)
         self._dbs = {None: weakref.ref(self._db)}
 
@@ -662,7 +673,8 @@ class Transaction(object):
 
         `write`:
             Transactions are read-only by default. To modify the database, you
-            must pass `write=True`.
+            must pass `write=True`. This flag is ignored if
+            :py:class:`Environment` was opened with ``readonly=True``.
 
         `buffers`:
             If ``True``, indicates :py:func:`buffer` objects should be yielded
@@ -685,6 +697,8 @@ class Transaction(object):
         self._to_py = _mvbuf if buffers else _mvstr
         self._deps = {}
         flags = 0
+        if write and env.readonly:
+            raise Error('Cannot start write transaction with read-only env')
         if not write:
             flags |= MDB_RDONLY
         txnpp = _ffi.new('MDB_txn **')
@@ -1086,6 +1100,14 @@ class Cursor(object):
         """
         return self._cursor_get_key(MDB_SET_KEY, key)
 
+    def get(self, key, default=None):
+        """Equivalent to :py:meth:`set_key()`, except :py:meth:`value` is
+        returned if `key` was found, otherwise `default`.
+        """
+        if self._cursor_get_key(MDB_SET_KEY, key):
+            return self.value()
+        return default
+
     def set_range(self, key):
         """Seek to the first key greater than or equal `key`, returning
         ``True`` on success, or ``False`` to indicate key was past end of
