@@ -20,11 +20,12 @@
  * <http://www.openldap.org/>.
  */
 
-#include <sys/stat.h>
-#include <stdbool.h>
-#include <string.h>
 #include <errno.h>
+#include <stdbool.h>
 #include <stdint.h>
+#include <string.h>
+#include <sys/stat.h>
+#include <tgmath.h>
 
 #include "Python.h"
 #include "structmember.h"
@@ -45,34 +46,77 @@
 #define NOINLINE __attribute__((noinline))
 
 static PyObject *Error;
-static PyObject *append_s;
-static PyObject *buffers_s;
-static PyObject *create_s;
-static PyObject *db_s;
-static PyObject *default_s;
-static PyObject *delete_s;
-static PyObject *dupdata_s;
-static PyObject *dupsort_s;
-static PyObject *key_s;
-static PyObject *keys_s;
-static PyObject *map_async_s;
-static PyObject *map_size_s;
-static PyObject *max_dbs_s;
-static PyObject *max_readers_s;
-static PyObject *metasync_s;
-static PyObject *mode_s;
-static PyObject *name_s;
-static PyObject *overwrite_s;
-static PyObject *parent_s;
-static PyObject *path_s;
-static PyObject *readonly_s;
-static PyObject *reverse_key_s;
-static PyObject *subdir_s;
-static PyObject *sync_s;
-static PyObject *txn_s;
-static PyObject *value_s;
-static PyObject *values_s;
-static PyObject *write_s;
+
+enum string_id {
+    APPEND_S,
+    BUFFERS_S,
+    CREATE_S,
+    DB_S,
+    DEFAULT_S,
+    DELETE_S,
+    DUPDATA_S,
+    DUPSORT_S,
+    KEY_S,
+    KEYS_S,
+    MAP_ASYNC_S,
+    MAP_SIZE_S,
+    MAX_DBS_S,
+    MAX_READERS_S,
+    METASYNC_S,
+    MODE_S,
+    NAME_S,
+    OVERWRITE_S,
+    PARENT_S,
+    PATH_S,
+    READONLY_S,
+    REVERSE_KEY_S,
+    SUBDIR_S,
+    SYNC_S,
+    TXN_S,
+    VALUE_S,
+    VALUES_S,
+    WRITE_S,
+    WRITEMAP_S,
+
+    // Must be last.
+    STRING_ID_COUNT
+};
+
+static const char *strings = (
+    "append\0"
+    "buffers\0"
+    "create\0"
+    "db\0"
+    "default\0"
+    "delete\0"
+    "dupdata\0"
+    "dupsort\0"
+    "key\0"
+    "keys\0"
+    "map_async\0"
+    "map_size\0"
+    "max_dbs\0"
+    "max_readers\0"
+    "metasync\0"
+    "mode\0"
+    "name\0"
+    "overwrite\0"
+    "parent\0"
+    "path\0"
+    "readonly\0"
+    "reverse_key\0"
+    "subdir\0"
+    "sync\0"
+    "txn\0"
+    "value\0"
+    "values\0"
+    "write\0"
+    "writemap\0"
+);
+static PyObject **string_tbl;
+static PyObject *py_zero;
+static PyObject *py_int_max;
+static PyObject *py_size_max;
 
 extern PyTypeObject PyDatabase_Type;
 extern PyTypeObject PyEnvironment_Type;
@@ -195,6 +239,7 @@ typedef struct EnvObject {
     LmdbObject_HEAD
     MDB_env *env;
     DbObject *main_db;
+    int readonly; // If 1, transactions are always readonly.
 } EnvObject;
 
 typedef struct {
@@ -363,13 +408,49 @@ type_error(const char *what)
 //-------------------------------
 
 #define OFFSET(k, y) offsetof(struct k, y)
-enum arg_type {ARG_EOF, ARG_BOOL, ARG_OBJ, ARG_BUF, ARG_STR, ARG_INT};
-struct argspec {
-    unsigned short type;
-    unsigned short offset;
-    PyObject **name;
-    PyTypeObject *objtype;
+#define SPECSIZE() (sizeof(argspec) / sizeof(argspec[0]))
+enum arg_type {
+    ARG_OBJ,
+    ARG_DB,
+    ARG_TRANS,
+    ARG_BOOL,
+    ARG_BUF,
+    ARG_STR,
+    ARG_INT,
+    ARG_SIZE
 };
+struct argspec {
+    unsigned char type;
+    unsigned char string_id;
+    unsigned short offset;
+};
+
+static PyTypeObject *type_tbl[] = {
+    &PyDatabase_Type,
+    &PyTransaction_Type,
+};
+
+
+static int
+parse_ulong(PyObject *obj, uint64_t *l, PyObject *max)
+{
+    int rc = PyObject_RichCompareBool(obj, py_zero, Py_GE);
+    if(rc == -1) {
+        return -1;
+    } else if(! rc) {
+        type_error("Integer argument must be >= 0");
+        return -1;
+    }
+    rc = PyObject_RichCompareBool(obj, max, Py_LE);
+    if(rc == -1) {
+        return -1;
+    } else if(! rc) {
+        type_error("Integer argument exceeds limit.");
+        return -1;
+    }
+    *l = PyInt_AsUnsignedLongLongMask(obj);
+    return 0;
+}
 
 
 static int
@@ -377,20 +458,22 @@ parse_arg(const struct argspec *spec, PyObject *val, void *out)
 {
     void *dst = ((uint8_t *)out) + spec->offset;
     int ret = 0;
+    uint64_t l;
 
     if(val != Py_None) {
-        switch(spec->type) {
-        case ARG_EOF:
-        case ARG_BOOL:
-            *((int *)dst) = val == Py_True;
-            break;
+        switch((enum arg_type) spec->type) {
         case ARG_OBJ:
-            if(spec->objtype && val->ob_type != spec->objtype) {
+        case ARG_DB:
+        case ARG_TRANS:
+            if(spec->type && val->ob_type != type_tbl[spec->type - 1]) {
                 type_error("invalid type");
                 return -1;
             }
             *((PyObject **) dst) = val;
             break;
+        case ARG_BOOL:
+            *((int *)dst) = val == Py_True;
+            break;
         case ARG_BUF:
             ret = val_from_buffer((MDB_val *)dst, val);
             break;
@@ -402,9 +485,17 @@ parse_arg(const struct argspec *spec, PyObject *val, void *out)
             break;
         }
         case ARG_INT:
-            *((int *) dst) = PyLong_AsUnsignedLong(val);
-            if(PyErr_Occurred()) {
+            if(parse_ulong(val, &l, py_int_max)) {
                 ret = -1;
+            } else {
+                *((int *) dst) = l;
+            }
+            break;
+        case ARG_SIZE:
+            if(parse_ulong(val, &l, py_size_max)) {
+                ret = -1;
+            } else {
+                *((size_t *) dst) = l;
             }
             break;
         }
@@ -418,7 +509,7 @@ parse_arg(const struct argspec *spec, PyObject *val, void *out)
  * module, keyword strings aren't dup'd every call and the code is >3x smaller.
  */
 static int NOINLINE
-parse_args(int valid, const struct argspec *argspec,
+parse_args(int valid, int specsize, const struct argspec *argspec,
            PyObject *args, PyObject *kwds, void *out)
 {
     if(! valid) {
@@ -427,35 +518,34 @@ parse_args(int valid, const struct argspec *argspec,
     }
 
     unsigned set = 0;
-    const struct argspec *spec = argspec;
     unsigned i;
     if(args) {
-        int pargsize = PyTuple_GET_SIZE(args);
-        i = 0;
-        while(i < pargsize && spec->type != ARG_EOF) {
-            if(parse_arg(spec, PyTuple_GET_ITEM(args, i), out)) {
-                return -1;
-            }
-            set |= (1 << i++);
-            spec++;
-        }
-
-        if(i >= pargsize && spec->type == ARG_EOF) {
+        int size = PyTuple_GET_SIZE(args);
+        if(size > specsize) {
             type_error("too many positional arguments.");
             return -1;
         }
+        size = fmin(specsize, size);
+        for(i = 0; i < size; i++) {
+            if(parse_arg(argspec + i, PyTuple_GET_ITEM(args, i), out)) {
+                return -1;
+            }
+            set |= 1 << i;
+        }
     }
 
     if(kwds) {
-        i = 0;
+        int size = PyDict_Size(kwds);
         int c = 0;
-        for(spec = argspec; spec->type != ARG_EOF; spec++) {
-            unsigned bit = 1 << i++;
-            PyObject *val = PyDict_GetItem(kwds, *spec->name);
+
+        for(i = 0; i < specsize && c != size; i++) {
+            const struct argspec *spec = argspec + i;
+            PyObject *kwd = string_tbl[spec->string_id];
+            PyObject *val = PyDict_GetItem(kwds, kwd);
             if(val) {
-                if(set & bit) {
+                if(set & (1 << i)) {
                     PyErr_Format(PyExc_TypeError, "duplicate argument: %s",
-                                 PyString_AS_STRING(*spec->name));
+                                 PyString_AS_STRING(kwd));
                     return -1;
                 }
                 if(parse_arg(spec, val, out)) {
@@ -465,7 +555,7 @@ parse_args(int valid, const struct argspec *argspec,
             }
         }
 
-        if(c != PyDict_Size(kwds)) {
+        if(c != size) {
             type_error("unrecognized keyword argument");
             return -1;
         }
@@ -510,7 +600,9 @@ txn_db_from_name(EnvObject *env, const char *name,
 {
     int rc;
     MDB_txn *txn;
-    if((rc = mdb_txn_begin(env->env, NULL, 0, &txn))) {
+
+    int begin_flags = (name == NULL || env->readonly) ? MDB_RDONLY : 0;
+    if((rc = mdb_txn_begin(env->env, NULL, begin_flags, &txn))) {
         err_set("mdb_txn_begin", rc);
         return NULL;
     }
@@ -583,12 +675,13 @@ env_dealloc(EnvObject *self)
     PyObject_Del(self);
 }
 
+
 static PyObject *
 env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
 {
     struct env_new {
         char *path;
-        int map_size;
+        size_t map_size;
         int subdir;
         int readonly;
         int metasync;
@@ -596,26 +689,27 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
         int map_async;
         int mode;
         int create;
+        int writemap;
         int max_readers;
         int max_dbs;
-    } arg = {NULL, 10485760, 1, 0, 1, 1, 0, 0644, 1, 126, 0};
+    } arg = {NULL, 10485760, 1, 0, 1, 1, 0, 0644, 1, 0, 126, 0};
 
     static const struct argspec argspec[] = {
-        {ARG_STR, OFFSET(env_new, path), &path_s, NULL},
-        {ARG_INT, OFFSET(env_new, map_size), &map_size_s, NULL},
-        {ARG_BOOL, OFFSET(env_new, subdir), &subdir_s, NULL},
-        {ARG_BOOL, OFFSET(env_new, readonly), &readonly_s, NULL},
-        {ARG_BOOL, OFFSET(env_new, metasync), &metasync_s, NULL},
-        {ARG_BOOL, OFFSET(env_new, sync), &sync_s, NULL},
-        {ARG_BOOL, OFFSET(env_new, map_async), &map_async_s, NULL},
-        {ARG_INT, OFFSET(env_new, mode), &mode_s, NULL},
-        {ARG_BOOL, OFFSET(env_new, create), &create_s, NULL},
-        {ARG_INT, OFFSET(env_new, max_readers), &max_readers_s, NULL},
-        {ARG_INT, OFFSET(env_new, max_dbs), &max_dbs_s, NULL},
-        {0, 0, 0, 0}
+        {ARG_STR, PATH_S, OFFSET(env_new, path)},
+        {ARG_SIZE, MAP_SIZE_S, OFFSET(env_new, map_size)},
+        {ARG_BOOL, SUBDIR_S, OFFSET(env_new, subdir)},
+        {ARG_BOOL, READONLY_S, OFFSET(env_new, readonly)},
+        {ARG_BOOL, METASYNC_S, OFFSET(env_new, metasync)},
+        {ARG_BOOL, SYNC_S, OFFSET(env_new, sync)},
+        {ARG_BOOL, MAP_ASYNC_S, OFFSET(env_new, map_async)},
+        {ARG_INT, MODE_S, OFFSET(env_new, mode)},
+        {ARG_BOOL, CREATE_S, OFFSET(env_new, create)},
+        {ARG_BOOL, WRITEMAP_S, OFFSET(env_new, writemap)},
+        {ARG_INT, MAX_READERS_S, OFFSET(env_new, max_readers)},
+        {ARG_INT, MAX_DBS_S, OFFSET(env_new, max_dbs)},
     };
 
-    if(parse_args(1, argspec, args, kwds, &arg)) {
+    if(parse_args(1, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
 
@@ -665,13 +759,14 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
         }
     }
 
-    int flags = 0; //MDB_WRITEMAP;
+    int flags = 0;
     if(! arg.subdir) {
         flags |= MDB_NOSUBDIR;
     }
     if(arg.readonly) {
         flags |= MDB_RDONLY;
     }
+    self->readonly = arg.readonly;
     if(! arg.metasync) {
         flags |= MDB_NOMETASYNC;
     }
@@ -681,6 +776,9 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     if(arg.map_async) {
         flags |= MDB_MAPASYNC;
     }
+    if(arg.writemap) {
+        flags |= MDB_WRITEMAP;
+    }
 
     DEBUG("mdb_env_open(%p, '%s', %d, %o);", self->env, arg.path, flags, arg.mode)
     if((rc = mdb_env_open(self->env, arg.path, flags, arg.mode))) {
@@ -721,12 +819,16 @@ make_trans(EnvObject *env, TransObject *parent, int write, int buffers)
         parent_txn = parent->txn;
     }
 
+    if(write && env->readonly) {
+        return err_set("Cannot start write transaction with read-only env", 0);
+    }
+
     TransObject *self = PyObject_New(TransObject, &PyTransaction_Type);
     if(! self) {
         return NULL;
     }
 
-    int flags = write ? 0 : MDB_RDONLY;
+    int flags = (write && !env->readonly) ? 0 : MDB_RDONLY;
     int rc = mdb_txn_begin(env->env, parent_txn, flags, &self->txn);
     if(rc) {
         PyObject_Del(self);
@@ -753,13 +855,12 @@ env_begin(EnvObject *self, PyObject *args, PyObject *kwds)
     } arg = {0, 0, NULL};
 
     static const struct argspec argspec[] = {
-        {ARG_BOOL, OFFSET(env_begin, buffers), &buffers_s, NULL},
-        {ARG_BOOL, OFFSET(env_begin, write), &write_s, NULL},
-        {ARG_OBJ, OFFSET(env_begin, parent), &parent_s, &PyTransaction_Type},
-        {0, 0, 0, 0}
+        {ARG_BOOL, BUFFERS_S, OFFSET(env_begin, buffers)},
+        {ARG_BOOL, WRITE_S, OFFSET(env_begin, write)},
+        {ARG_TRANS, PARENT_S, OFFSET(env_begin, parent)},
     };
 
-    if(parse_args(self->valid, argspec, args, kwds, &arg)) {
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
     return make_trans(self, arg.parent, arg.write, arg.buffers);
@@ -831,15 +932,14 @@ env_open_db(EnvObject *self, PyObject *args, PyObject *kwds)
     } arg = {NULL, NULL, 0, 0, 1};
 
     static const struct argspec argspec[] = {
-        {ARG_STR, OFFSET(env_open_db, name), &name_s, NULL},
-        {ARG_OBJ, OFFSET(env_open_db, txn), &txn_s, &PyTransaction_Type},
-        {ARG_BOOL, OFFSET(env_open_db, reverse_key), &reverse_key_s, NULL},
-        {ARG_BOOL, OFFSET(env_open_db, dupsort), &dupsort_s, NULL},
-        {ARG_BOOL, OFFSET(env_open_db, create), &create_s, NULL},
-        {0, 0, 0, 0}
+        {ARG_STR, NAME_S, OFFSET(env_open_db, name)},
+        {ARG_TRANS, TXN_S, OFFSET(env_open_db, txn)},
+        {ARG_BOOL, REVERSE_KEY_S, OFFSET(env_open_db, reverse_key)},
+        {ARG_BOOL, DUPSORT_S, OFFSET(env_open_db, dupsort)},
+        {ARG_BOOL, CREATE_S, OFFSET(env_open_db, create)},
     };
 
-    if(parse_args(1, argspec, args, kwds, &arg)) {
+    if(parse_args(1, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
 
@@ -1027,12 +1127,11 @@ cursor_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     } arg = {NULL, NULL};
 
     static const struct argspec argspec[] = {
-        {ARG_OBJ, OFFSET(cursor_new, db), &db_s, &PyDatabase_Type},
-        {ARG_OBJ, OFFSET(cursor_new, trans), &txn_s, &PyTransaction_Type},
-        {0, 0, 0, 0}
+        {ARG_DB, DB_S, OFFSET(cursor_new, db)},
+        {ARG_TRANS, TXN_S, OFFSET(cursor_new, trans)}
     };
 
-    if(parse_args(1, argspec, args, kwds, &arg)) {
+    if(parse_args(1, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
 
@@ -1121,6 +1220,48 @@ cursor_first(CursorObject *self)
     return _cursor_get(self, MDB_FIRST);
 }
 
+
+static PyObject *
+cursor_value(CursorObject *self);
+
+
+static PyObject *
+cursor_get(CursorObject *self, PyObject *args, PyObject *kwds)
+{
+    if(! self->valid) {
+        return err_invalid();
+    }
+
+    struct cursor_get {
+        MDB_val key;
+        PyObject *default_;
+    } arg = {{0, 0}, Py_None};
+
+    static const struct argspec argspec[] = {
+        {ARG_BUF, KEY_S, OFFSET(cursor_get, key)},
+        {ARG_OBJ, DEFAULT_S, OFFSET(cursor_get, default_)}
+    };
+
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
+        return NULL;
+    }
+
+    if(! arg.key.mv_data) {
+        return type_error("key must be given.");
+    }
+
+    self->key = arg.key;
+    if(_cursor_get_c(self, MDB_SET_KEY)) {
+        return NULL;
+    }
+    if(! self->positioned) {
+        Py_INCREF(arg.default_);
+        return arg.default_;
+    }
+    return cursor_value(self);
+}
+
+
 static PyObject *
 cursor_item(CursorObject *self)
 {
@@ -1214,15 +1355,14 @@ cursor_put(CursorObject *self, PyObject *args, PyObject *kwds)
     } arg = {{0, 0}, {0, 0}, 0, 1, 0};
 
     static const struct argspec argspec[] = {
-        {ARG_BUF, OFFSET(cursor_put, key), &key_s, NULL},
-        {ARG_BUF, OFFSET(cursor_put, val), &value_s, NULL},
-        {ARG_BOOL, OFFSET(cursor_put, dupdata), &dupdata_s, NULL},
-        {ARG_BOOL, OFFSET(cursor_put, overwrite), &overwrite_s, NULL},
-        {ARG_BOOL, OFFSET(cursor_put, append), &append_s, NULL},
-        {0, 0, 0, 0}
+        {ARG_BUF, KEY_S, OFFSET(cursor_put, key)},
+        {ARG_BUF, VALUE_S, OFFSET(cursor_put, val)},
+        {ARG_BOOL, DUPDATA_S, OFFSET(cursor_put, dupdata)},
+        {ARG_BOOL, OVERWRITE_S, OFFSET(cursor_put, overwrite)},
+        {ARG_BOOL, APPEND_S, OFFSET(cursor_put, append)}
     };
 
-    if(parse_args(self->valid, argspec, args, kwds, &arg)) {
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
 
@@ -1305,12 +1445,11 @@ iter_from_args(CursorObject *self, PyObject *args, PyObject *kwds,
     } arg = {1, 1};
 
     static const struct argspec argspec[] = {
-        {ARG_BOOL, OFFSET(iter_from_args, keys), &keys_s, NULL},
-        {ARG_BOOL, OFFSET(iter_from_args, values), &values_s, NULL},
-        {0, 0, 0, 0}
+        {ARG_BOOL, KEYS_S, OFFSET(iter_from_args, keys)},
+        {ARG_BOOL, VALUES_S, OFFSET(iter_from_args, values)}
     };
 
-    if(parse_args(self->valid, argspec, args, kwds, &arg)) {
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
 
@@ -1362,6 +1501,7 @@ static struct PyMethodDef cursor_methods[] = {
     {"count", (PyCFunction)cursor_count, METH_NOARGS},
     {"delete", (PyCFunction)cursor_delete, METH_NOARGS},
     {"first", (PyCFunction)cursor_first, METH_NOARGS},
+    {"get", (PyCFunction)cursor_get, METH_VARARGS|METH_KEYWORDS},
     {"item", (PyCFunction)cursor_item, METH_NOARGS},
     {"iternext", (PyCFunction)cursor_iternext, METH_VARARGS|METH_KEYWORDS},
     {"iterprev", (PyCFunction)cursor_iterprev, METH_VARARGS|METH_KEYWORDS},
@@ -1494,17 +1634,16 @@ trans_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     } arg = {NULL, NULL, 0, 0};
 
     static const struct argspec argspec[] = {
-        {ARG_OBJ, OFFSET(trans_new, env), &txn_s, &PyTransaction_Type},
-        {ARG_OBJ, OFFSET(trans_new, parent), &parent_s, &PyTransaction_Type},
-        {ARG_BOOL, OFFSET(trans_new, write), &write_s, NULL},
-        {ARG_BOOL, OFFSET(trans_new, buffers), &buffers_s, NULL},
-        {0, 0, 0, 0}
+        {ARG_TRANS, TXN_S, OFFSET(trans_new, env)},
+        {ARG_TRANS, PARENT_S, OFFSET(trans_new, parent)},
+        {ARG_BOOL, WRITE_S, OFFSET(trans_new, write)},
+        {ARG_BOOL, BUFFERS_S, OFFSET(trans_new, buffers)}
     };
 
     if(! arg.env) {
         return type_error("'env' argument required");
     }
-    if(parse_args(1, argspec, args, kwds, &arg)) {
+    if(parse_args(1, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
     return make_trans(arg.env, arg.parent, arg.write, arg.buffers);
@@ -1551,11 +1690,10 @@ trans_cursor(TransObject *self, PyObject *args, PyObject *kwds)
     } arg = { 0 };
 
     static const struct argspec argspec[] = {
-        {ARG_OBJ, OFFSET(trans_cursor, db), &db_s, &PyDatabase_Type},
-        {0, 0, 0, 0}
+        {ARG_DB, DB_S, OFFSET(trans_cursor, db)}
     };
 
-    if(parse_args(self->valid, argspec, args, kwds, &arg)) {
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
     if(! arg.db) {
@@ -1575,13 +1713,12 @@ trans_delete(TransObject *self, PyObject *args, PyObject *kwds)
     } arg = {{0, 0}, {0, 0}, NULL};
 
     static const struct argspec argspec[] = {
-        {ARG_BUF, OFFSET(trans_delete, key), &key_s, NULL},
-        {ARG_BUF, OFFSET(trans_delete, val), &value_s, NULL},
-        {ARG_OBJ, OFFSET(trans_delete, db), &db_s, &PyDatabase_Type},
-        {0, 0, 0, 0}
+        {ARG_BUF, KEY_S, OFFSET(trans_delete, key)},
+        {ARG_BUF, VALUE_S, OFFSET(trans_delete, val)},
+        {ARG_DB, DB_S, OFFSET(trans_delete, db)}
     };
 
-    if(parse_args(self->valid, argspec, args, kwds, &arg)) {
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
     if(! arg.db) {
@@ -1608,12 +1745,11 @@ trans_drop(TransObject *self, PyObject *args, PyObject *kwds)
     } arg = { NULL, 1 };
 
     static const struct argspec argspec[] = {
-        {ARG_OBJ, OFFSET(trans_drop, db), &db_s, &PyDatabase_Type},
-        {ARG_BOOL, OFFSET(trans_drop, delete), &delete_s, NULL},
-        {0, 0, 0, 0}
+        {ARG_DB, DB_S, OFFSET(trans_drop, db)},
+        {ARG_BOOL, DELETE_S, OFFSET(trans_drop, delete)}
     };
 
-    if(parse_args(self->valid, argspec, args, kwds, &arg)) {
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
     if(! arg.db) {
@@ -1637,13 +1773,12 @@ trans_get(TransObject *self, PyObject *args, PyObject *kwds)
     } arg = {{0, 0}, Py_None, self->env->main_db};
 
     static const struct argspec argspec[] = {
-        {ARG_BUF, OFFSET(trans_get, key), &key_s, NULL},
-        {ARG_OBJ, OFFSET(trans_get, default_), &default_s, NULL},
-        {ARG_OBJ, OFFSET(trans_get, db), &db_s, &PyDatabase_Type},
-        {0, 0, 0, 0}
+        {ARG_BUF, KEY_S, OFFSET(trans_get, key)},
+        {ARG_OBJ, DEFAULT_S, OFFSET(trans_get, default_)},
+        {ARG_DB, DB_S, OFFSET(trans_get, db)}
     };
 
-    if(parse_args(self->valid, argspec, args, kwds, &arg)) {
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
 
@@ -1679,16 +1814,15 @@ trans_put(TransObject *self, PyObject *args, PyObject *kwds)
     } arg = {{0, 0}, {0, 0}, 0, 1, 0, self->env->main_db};
 
     static const struct argspec argspec[] = {
-        {ARG_BUF, OFFSET(trans_put, key), &key_s, NULL},
-        {ARG_BUF, OFFSET(trans_put, value), &value_s, NULL},
-        {ARG_BOOL, OFFSET(trans_put, dupdata), &dupdata_s, NULL},
-        {ARG_BOOL, OFFSET(trans_put, overwrite), &overwrite_s, NULL},
-        {ARG_BOOL, OFFSET(trans_put, append), &append_s, NULL},
-        {ARG_OBJ, OFFSET(trans_put, db), &db_s, &PyDatabase_Type},
-        {0, 0, 0, 0}
+        {ARG_BUF, KEY_S, OFFSET(trans_put, key)},
+        {ARG_BUF, VALUE_S, OFFSET(trans_put, value)},
+        {ARG_BOOL, DUPDATA_S, OFFSET(trans_put, dupdata)},
+        {ARG_BOOL, OVERWRITE_S, OFFSET(trans_put, overwrite)},
+        {ARG_BOOL, APPEND_S, OFFSET(trans_put, append)},
+        {ARG_DB, DB_S, OFFSET(trans_put, db)}
     };
 
-    if(parse_args(self->valid, argspec, args, kwds, &arg)) {
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
 
@@ -1797,41 +1931,27 @@ initcpython(void)
         }
     }
 
-    static struct { PyObject **obj; const char *s; } strs[] = {
-        {&append_s, "append"},
-        {&buffers_s, "buffers"},
-        {&create_s, "create"},
-        {&db_s, "db"},
-        {&default_s, "default"},
-        {&dupdata_s, "dupdata"},
-        {&dupsort_s, "dupsort"},
-        {&key_s, "key"},
-        {&keys_s, "keys"},
-        {&map_async_s, "map_async"},
-        {&map_size_s, "map_size"},
-        {&max_dbs_s, "max_dbs"},
-        {&max_readers_s, "max_readers"},
-        {&metasync_s, "metasync"},
-        {&mode_s, "mode"},
-        {&name_s, "name"},
-        {&overwrite_s, "overwrite"},
-        {&parent_s, "parent"},
-        {&path_s, "path"},
-        {&readonly_s, "readonly"},
-        {&reverse_key_s, "reverse_key"},
-        {&subdir_s, "subdir"},
-        {&sync_s, "sync"},
-        {&txn_s, "txn"},
-        {&value_s, "value"},
-        {&values_s, "values"},
-        {&write_s, "write"},
-        {NULL, NULL}
-    };
-    for(i = 0; strs[i].obj; i++) {
-        *strs[i].obj = PyString_InternFromString(strs[i].s);
-        if(! *strs[i].obj) {
+    string_tbl = malloc(sizeof(PyObject *) * STRING_ID_COUNT);
+    if(! string_tbl) {
+        return;
+    }
+
+    const char *cur = strings;
+    for(i = 0; i < STRING_ID_COUNT; i++) {
+        if(! ((string_tbl[i] = PyString_InternFromString(cur)))) {
             return;
         }
+        cur += strlen(cur) + 1;
+    }
+
+    if(! ((py_zero = PyLong_FromSize_t(0)))) {
+        return;
+    }
+    if(! ((py_int_max = PyLong_FromSize_t(INT_MAX)))) {
+        return;
+    }
+    if(! ((py_size_max = PyLong_FromSize_t(SIZE_MAX)))) {
+        return;
     }
 
     Error = PyErr_NewException("lmdb.Error", NULL, NULL);
